#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include "mt64.h"
#include "percolating_clusters.h"



#define SENS 1e-7


/////////////////////////////////////////////////////////////////
void iterate_equation_infection_nodes(int N, int M, int **ref_vec, int **intersection, int **reference_int, int **onlyA, int **reference_A, int **onlyB, int **reference_B, double *probabilities, double *old_probabilities, double p)
{

  double I, A, B;
  int i, j, n, s, n1, n2, m;

  for(i=1;i<=M;i++) old_probabilities[i] = probabilities[i];


  for(i=1;i<=M;i++)
    {
      I = A = B = 1.0;

      n1 = ref_vec[i][0];
      n2 = ref_vec[i][1];


      for(j=1;j<=intersection[0][n1];j++)
        {
          n = intersection[n1][j];
          if (n != n2)
            {
              m = reference_int[n1][j];
              I = I * (1.0 - old_probabilities[m]);
            }
        }


      for(j=1;j<=onlyA[0][n1];j++)
        {
          n = onlyA[n1][j];
          if (n != n2)
            {
              m = reference_A[n1][j];
              A = A * (1.0 - old_probabilities[m]);
            }
        }


      for(j=1;j<=onlyB[0][n1];j++)
	{
          n = onlyB[n1][j];
          if (n != n2)
            {
              m = reference_B[n1][j];
              B = B * (1.0 - old_probabilities[m]);
            }
        }

      //printf("%g %g %g %g\n",p,I,A,B);

      A = 1.0 - A;
      B = 1.0 - B;
      I = 1.0 - I;
      
      probabilities[i] = p * ( I + (1.0-I)*A*B );

    }

}



void put_together_infection_nodes(double *vec, int N, int M, int **ref_vec, int **intersection, int **reference_int, int **onlyA, int **reference_A, int **onlyB, int **reference_B, double *probabilities, double p)
{

  double I, A, B;
  int i, j, n, r;


  for(i=1;i<=N;i++)                                                                                                                                                                
    {                                                                                                                                                                              
      I = A = B = 1.0; 

      for(j=1;j<=intersection[0][i];j++)                                                                                                                                   
   	{                                                                                                                                                                       
	  n = intersection[i][j];
	  r = reference_int[i][j];	  
	  I = I * (1.0 - probabilities[r]);                                                                                                                           
	} 

      for(j=1;j<=onlyA[0][i];j++)
        {
          n = onlyA[i][j];
          r = reference_A[i][j];
          A = A * (1.0 - probabilities[r]);
        }


      for(j=1;j<=onlyB[0][i];j++)
        {
          n = onlyB[i][j];
          r = reference_B[i][j];
          B = B * (1.0 - probabilities[r]);
        }

      I = 1.0 - I;
      A = 1.0 - A;
      B = 1.0 - B;

      vec[i] = p *(I + (1.0 - I) *A *B);

    }       



}



void theoretical_model_infection_nodes (int N, int **intersection, int **onlyA, int **onlyB, double **order, int V)
{
  double err, p;
  int i, j, m, n, s, r;

  int M_int = 0;
  for(i=1;i<=N;i++) M_int+=intersection[0][i];
  int M_A = 0;
  for(i=1;i<=N;i++) M_A+=onlyA[0][i];
  int M_B = 0;
  for(i=1;i<=N;i++) M_B+=onlyB[0][i];


  int M = M_int + M_A + M_B;


  int **ref_vec = (int **)malloc((M+1)*sizeof(int *));
  for(i=1;i<=M;i++) ref_vec[i] = (int *)malloc(2*sizeof(int));
  double *probabilities = (double *)malloc((M+1)*sizeof(double));
  double *old_probabilities = (double *)malloc((M+1)*sizeof(double));


  double *vec = (double *)malloc((N+1)*sizeof(double));




  m = 0;


  //intersection
  //memory allocation                                                                                                                   
  int **reference_int = (int **)malloc((N+1)*sizeof(int *));
  for(i=1;i<=N;i++) reference_int[i] = (int *)malloc((intersection[0][i]+1)*sizeof(int));
  /////             
  m = 0;
  for(i=1;i<=N;i++)
    {
      for(j=1;j<=intersection[0][i];j++)
        {
          n = intersection[i][j];
          m++;
          ref_vec[m][0] = i;
          ref_vec[m][1] = n;
          s = find_node(n, i, intersection);
          reference_int[n][s] = m;
        }
    }
  ////////////////////////////////////////



  //onlyA                           
  //memory allocation                                                                                                                   
  int **reference_A = (int **)malloc((N+1)*sizeof(int *));
  for(i=1;i<=N;i++) reference_A[i] = (int *)malloc((onlyA[0][i]+1)*sizeof(int));
  /////                                                                                                                                 

  for(i=1;i<=N;i++)
    {
      for(j=1;j<=onlyA[0][i];j++)
        {
          n = onlyA[i][j];
          m++;
          ref_vec[m][0] = i;
          ref_vec[m][1] = n;
          s = find_node(n, i, onlyA);
          reference_A[n][s] = m;
        }
    }
  //////////////////////////////////////// 



  //onlyB                                                                                                                               
  //memory allocation                                                                                                                   
  int **reference_B = (int **)malloc((N+1)*sizeof(int *));
  for(i=1;i<=N;i++) reference_B[i] = (int *)malloc((onlyB[0][i]+1)*sizeof(int));
  /////                                                                                                                                 
  for(i=1;i<=N;i++)
    {
      for(j=1;j<=onlyB[0][i];j++)
        {
          n = onlyB[i][j];
          m++;
          ref_vec[m][0] = i;
          ref_vec[m][1] = n;
          s = find_node(n, i, onlyB);
          reference_B[n][s] = m;
	}
    }
  ////////////////////////////////////////   









  for(i=0;i<=V;i++)
    {

      p = order[i][0];
      printf("#%g\n",p);


      //for(j=1;j<=M;j++) probabilities[j] = genrand64_real3();
      for(j=1;j<=M;j++) probabilities[j] = 1.0;
  
      err = 1.0e10;
      while(err>SENS)
	{

	  iterate_equation_infection_nodes(N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, old_probabilities, p);
	  err = estimate_error_basic (M, probabilities, old_probabilities);
	}
  
      put_together_infection_nodes(vec, N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, p);

      order[i][1] = 0.0;
      for(j=1;j<=N;j++) order[i][1] += vec[j];
      order[i][1] /= (double)N;
    }










  //intersection
  for(i=1;i<=N;i++) free(reference_int[i]);
  free(reference_int);
  //onlyA                                                                                                                   
  for(i=1;i<=N;i++) free(reference_A[i]);
  free(reference_A);
  //onlyB                                                                                                                              
  for(i=1;i<=N;i++) free(reference_B[i]);
  free(reference_B);



  for(i=1;i<=M;i++) free(ref_vec[i]);
  free(ref_vec);
  free(probabilities);
  free(old_probabilities);
  free(vec);

}




void find_transition_theoretical_model_infection_nodes (int N, int **intersection, int **onlyA, int **onlyB, double *res)
{
  double err;
  int i, j, m, n, s, r;

  int M_int = 0;
  for(i=1;i<=N;i++) M_int+=intersection[0][i];
  int M_A = 0;
  for(i=1;i<=N;i++) M_A+=onlyA[0][i];
  int M_B = 0;
  for(i=1;i<=N;i++) M_B+=onlyB[0][i];


  int M = M_int + M_A + M_B;


  int **ref_vec = (int **)malloc((M+1)*sizeof(int *));
  for(i=1;i<=M;i++) ref_vec[i] = (int *)malloc(2*sizeof(int));
  double *probabilities = (double *)malloc((M+1)*sizeof(double));
  double *old_probabilities = (double *)malloc((M+1)*sizeof(double));


  double *vec = (double *)malloc((N+1)*sizeof(double));




  m = 0;


  //intersection
  //memory allocation                                                                                                                   
  int **reference_int = (int **)malloc((N+1)*sizeof(int *));
  for(i=1;i<=N;i++) reference_int[i] = (int *)malloc((intersection[0][i]+1)*sizeof(int));
  /////             
  m = 0;
  for(i=1;i<=N;i++)
    {
      for(j=1;j<=intersection[0][i];j++)
        {
          n = intersection[i][j];
          m++;
          ref_vec[m][0] = i;
          ref_vec[m][1] = n;
          s = find_node(n, i, intersection);
          reference_int[n][s] = m;
        }
    }
  ////////////////////////////////////////



  //onlyA                           
  //memory allocation                                                                                                                   
  int **reference_A = (int **)malloc((N+1)*sizeof(int *));
  for(i=1;i<=N;i++) reference_A[i] = (int *)malloc((onlyA[0][i]+1)*sizeof(int));
  /////                                                                                                                                 

  for(i=1;i<=N;i++)
    {
      for(j=1;j<=onlyA[0][i];j++)
        {
          n = onlyA[i][j];
          m++;
          ref_vec[m][0] = i;
          ref_vec[m][1] = n;
          s = find_node(n, i, onlyA);
          reference_A[n][s] = m;
        }
    }
  //////////////////////////////////////// 



  //onlyB                                                                                                                               
  //memory allocation                                                                                                                   
  int **reference_B = (int **)malloc((N+1)*sizeof(int *));
  for(i=1;i<=N;i++) reference_B[i] = (int *)malloc((onlyB[0][i]+1)*sizeof(int));
  /////                                                                                                                                 
  for(i=1;i<=N;i++)
    {
      for(j=1;j<=onlyB[0][i];j++)
        {
          n = onlyB[i][j];
          m++;
          ref_vec[m][0] = i;
          ref_vec[m][1] = n;
          s = find_node(n, i, onlyB);
          reference_B[n][s] = m;
	}
    }
  ////////////////////////////////////////   




  double p1, p2, p;
  double order1, order2, order;
  double dp = 1e-4;

  p1 = dp;
  p2 = 1.0 -dp;

  while(p2-p1>dp){

    p = (p2+p1)/2.0;

    
    ////////
    //p1
    //for(j=1;j<=M;j++) probabilities[j] = genrand64_real3();  
    for(j=1;j<=M;j++) probabilities[j] = 1.0;
    err = 1.0e10;
    while(err>SENS)
      {
	iterate_equation_infection_nodes(N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, old_probabilities, p1);
	err = estimate_error_basic (M, probabilities, old_probabilities);
      }
    put_together_infection_nodes(vec, N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, p1);

    order1 = 0.0;
    for(j=1;j<=N;j++) order1 += vec[j];
    order1 /= (double)N;



    //p2
    //
    //for(j=1;j<=M;j++) probabilities[j] = genrand64_real3();  
    for(j=1;j<=M;j++) probabilities[j] = 1.0;
    err = 1.0e10;
    while(err>SENS)
      {
	iterate_equation_infection_nodes(N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, old_probabilities, p2);
	err = estimate_error_basic (M, probabilities, old_probabilities);
      }
    put_together_infection_nodes(vec, N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, p2);

    order2 = 0.0;
    for(j=1;j<=N;j++) order2 += vec[j];
    order2 /= (double)N;
    

    //p 
    //
    //for(j=1;j<=M;j++) probabilities[j] = genrand64_real3();  
    for(j=1;j<=M;j++) probabilities[j] = 1.0;
    err = 1.0e10;
    while(err>SENS)
      {
	iterate_equation_infection_nodes(N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, old_probabilities, p);
	err = estimate_error_basic (M, probabilities, old_probabilities);
      }
    put_together_infection_nodes(vec, N, M, ref_vec, intersection, reference_int, onlyA, reference_A, onlyB, reference_B, probabilities, p);

    order = 0.0;
    for(j=1;j<=N;j++) order += vec[j];
    order /= (double)N;



    printf("#%g %g\t%g %g\t%g %g\n",p1,order1,p2,order2,p,order);

    
    if(order1>1.0/(double)N && order2>1.0/(double)N)
      {
	printf("#Error A\n");
      }

    if(order1<1.0/(double)N && order2<1.0/(double)N)
      {
	printf("#Error B\n");
      }


    if(order1 < 1.0/(double)N && order < 1.0 /(double)N)
      {
	p1 = p;
      }

    if(order2 > 1.0/(double)N && order > 1.0 /(double)N)
      {
	p2 = p;
      }

      
  }


  res[0] = order2 - order1;
  res[1] = p;








  //intersection
  for(i=1;i<=N;i++) free(reference_int[i]);
  free(reference_int);
  //onlyA                                                                                                                   
  for(i=1;i<=N;i++) free(reference_A[i]);
  free(reference_A);
  //onlyB                                                                                                                              
  for(i=1;i<=N;i++) free(reference_B[i]);
  free(reference_B);



  for(i=1;i<=M;i++) free(ref_vec[i]);
  free(ref_vec);
  free(probabilities);
  free(old_probabilities);
  free(vec);

}




//////////////////////////////////////////////////////////////////

double estimate_error_basic (int N, double *vec1, double *vec2)
{
  int i;
  double err = 0.0;
  for(i=1;i<=N;i++) err += (vec1[i]-vec2[i])* (vec1[i]-vec2[i]);
  err = sqrt(err)/(double)N;
  return err;
}


///////////////////////////////////////////////////////
//old functions, defined on adjacency matrices and valid in absence of intersection

void  iterate_equation_nodes_multiplex (int N, int **bond1, int **bond2, double *vec1, double *vec2, double p)
{
  int i, j, n;
  double a1, a2;
  for(i=1;i<=N;i++)
    {
      vec2[i] = vec1[i];
    }
  for(i=1;i<=N;i++)
    {
      a1=a2=1.0;
      for(j=1;j<=bond1[0][i];j++) a1 = a1 * (1.0 - vec2[bond1[i][j]]);
      for(j=1;j<=bond2[0][i];j++) a2 = a2 * (1.0 - vec2[bond2[i][j]]);
      vec1[i] = 1.0 - ((1.0-a1)*(1.0-a2));
      vec1[i] = p - p*vec1[i];
    }
}


double theoretical_model_nodes_multiplex(int N, int **bond1, int **bond2, double p)
{
  double err, a1, a2, order;
  int i, m, j;

  double *vec1 = (double *)malloc((N+1)*sizeof(double));
  double *vec2 = (double *)malloc((N+1)*sizeof(double));


  //for(i=1;i<=N;i++) vec1[i] = genrand64_real3();
 for(i=1;i<=N;i++) vec1[i] = 1.0;
  err = 1.0e10;
  while(err>SENS){
    iterate_equation_nodes_multiplex (N, bond1, bond2, vec1, vec2, p);
    err = estimate_error_basic (N, vec1, vec2);
  }



  order = 0.0;
  for(i=1;i<=N;i++) order += vec1[i];
  order /= (double)N;



    


  free(vec1);
  free(vec2);

  return order;
}



















///////////////////////////////////////////////////////
int find_node (int i, int j, int **bond)
{
  int k;
  for(k=1;k<=bond[0][i];k++) if(bond[i][k]==j) return k;
  return -1;
}

void add_edge (int i, int n, int **bond)
{
  bond[0][i] ++;
  bond[i] = (int *)realloc(bond[i],(bond[0][i]+1)*sizeof(int));
  bond[i][bond[0][i]] = n;
  bond[0][n] ++;
  bond[n] = (int *)realloc(bond[n],(bond[0][n]+1)*sizeof(int));
  bond[n][bond[0][n]] = i;
}


void find_intersection (int N, int **bondA, int **bondB, int **intersection, int **onlyA, int **onlyB)
{

  int i, j, n, l;

  for(i=1;i<=N;i++)
    {
      intersection[0][i] = onlyA[0][i] = onlyB[0][i] =0;
    }


  for(i=1;i<=N;i++)
    {
      for(j=1;j<=bondA[0][i];j++)
	{
	  n = bondA[i][j];
	  if(n>i){
	    l = find_node(i, n, bondB);
	    if(l<0) add_edge(i, n, onlyA);
	    else add_edge(i, n, intersection);
	  }
	}
    }

      
  for(i=1;i<=N;i++)
    {
      for(j=1;j<=bondB[0][i];j++)
	{
	  n = bondB[i][j];
	  if(n>i){
	    l = find_node(i, n, bondA);
	    if(l<0) add_edge(i,n, onlyB);
	  }
	}
    }


}














////////////////////////////////////////////////////////


int mutually_connected_largest (int N, int **bondA, int **bondB, double p)
{

  int i, old;
  int *present = (int *)malloc((N+1)*sizeof(int));
  int *clusterA = (int *)malloc((N+1)*sizeof(int));
  int *clusterB = (int *)malloc((N+1)*sizeof(int));
  
  present[0] = 0;
  for(i=1;i<=N;i++)  
    {
      present[i] = 0;
      if (genrand64_real3()<p) 
	{
	  present[i] = 1;
	  present[0] ++;
	}
    }


  old = present[0];




  check_component (N, bondA, clusterA, present);
  check_component (N, bondB, clusterB, present);

  present[0] = 0;

  for(i=1;i<=N;i++) 
    {
      present[i] = clusterA[i]*clusterB[i];
      present[0] += present[i];
    }

  while(old!=present[0])
    {
      old = present[0];
      check_component (N, bondA, clusterA, present);
      check_component (N, bondB, clusterB, present);
      present[0] = 0;
      for(i=1;i<=N;i++) 
	{
	  present[i] = clusterA[i]*clusterB[i];
	  present[0] += present[i];
	}
    }






  free(present);
  free(clusterA);
  free(clusterB);

  return old;
}










void check_component (int N, int **bond, int *cluster, int *occupied)
{
  int i, j, k, s, n;

  int *comp=(int *)malloc((N+1)*sizeof(int));
  int *vec1=(int *)malloc((N+1)*sizeof(int));
  int *vec2=(int *)malloc((N+1)*sizeof(int));

  for(i=0;i<=N;i++)  comp[i]=-1;
  comp[0] = 0;


  for(n=1;n<=N;n++)
    {

      if(comp[n]<0 && occupied[n]>0){

        s=n;

        comp[0]++;
        comp[s] = comp[0];

        vec1[0] = 1;
        vec1[1] = s;


	while(vec1[0]>0) {

          vec2[0]=0;
          for(i=1;i<=vec1[0];i++)
            {
              s=vec1[i];


              for(j=1;j<=bond[0][s];j++)
                {
                  k=bond[s][j];
                  if(comp[k]<0 && occupied[k]>0)
                    {
                      vec2[0]++;
                      vec2[vec2[0]]=k;
                      comp[k]=comp[0];
                    }
                }

            }

	  vec1[0]=vec2[0];
	  if(vec2[0]>0) for(i=1;i<=vec2[0];i++) vec1[i]=vec2[i];
        }
      }
    }


  for(i=1;i<=N;i++) cluster[i] = 0;
  for(i=1;i<=N;i++) 
    {
      if(comp[i]>0){
	cluster[comp[i]]++;
      }
    }
  int max_c, max_s;

  max_c = -100;
  max_s = 0;
  for(i=1;i<=N;i++) 
    {
      if(max_s<cluster[i])
	{
	  max_s = cluster[i];
	  max_c = i;
	}
    }


  for(i=1;i<=N;i++) 
    {
      cluster[i] = 0;
      if(comp[i] == max_c) cluster[i] = 1;
    }



  free(comp);
  free(vec1);
  free(vec2);
}
